from django.contrib import admin
from django.http import HttpResponseRedirect

from .models import (
    Address,
    Job,
    JobNote,
    JobDate,
    Construction,
    Submittal,
    Stair,
    Company,
    ProjectManager,
    Contact,
    Transmittal,
    TransmittalOverride,
    Task,
    )

admin.site.site_header = 'Drawing Status admin'
admin.site.site_title = 'Drawing Status'


class SaveViewMixin:
    def response_change(self, request, obj):
        if "_view" in request.POST:
            return HttpResponseRedirect(obj.get_absolute_url())
        else:
            return super().response_change(request, obj)

    def response_add(self, request, obj):
        if "_view" in request.POST:
            return HttpResponseRedirect(obj.get_absolute_url())
        else:
            return super().response_add(request, obj)


class SubInline(admin.TabularInline):
    show_change_link = True
    model = Submittal
    extra = 1


class JobNoteInline(admin.TabularInline):
    model = JobNote
    extra = 0


class JobDateInline(admin.TabularInline):
    model = JobDate
    extra = 0


class ConstructionInline(admin.TabularInline):
    model = Construction
    extra = 0


@admin.action(description='Archive selected jobs')
def archive_jobs(modeladmin, request, queryset):
    queryset.update(archived=True)

@admin.register(Job)
class JobAdmin(SaveViewMixin, admin.ModelAdmin):
    def get_queryset(self, request):
        return Job.objects.all()

    actions = [archive_jobs]
    list_display = [
        'number',
        'name',
        'sub_detailers',
        'fab_detailers',
        # 'stairs_in_design',
        # 'stairs_in_fab',
        # 'stairs_at_engineering',
        # 'stairs_at_submittal',
    ]
    list_display_links = ['name']
    list_filter = [
        'task__detailer',
        'construction__style',
        'submittal__stair__stringer_type',
        'submittal__stair__stair_con',
        'submittal__stair__land_con',
        'submittal__stair__rail_type',
        'submittal__stair__finish',
    ]
    search_fields = [
        'number',
        'name',
        'submittal__stair__finish',
        'submittal__stair__finish_custom',
        'submittal__stair__stn',
        'submittal__stair__stn_note',
        'submittal__stair__stringer_type',
        'addresses__name',
        'addresses__street_address',
        'addresses__city',
        'addresses__state',
        'addresses__zip_code',
        'construction__style',
        'construction__style_custom',
        'project_manager__name',
        'contractor__name',
        'architect__name',
        'eor__name',
        'customer__name',
        'task__detailer__first_name',
    ]
    inlines = [ConstructionInline, JobDateInline, JobNoteInline, SubInline]
    fieldsets = [
        (None, {
            'fields': [
                ('number',
                 'name',
                 'award_date',),
                'addresses',
                'archived',
            ]
        }),
        ('Client Info', {
            'fields': [
                'contractors',
                'architects',
                'eors',
                'customers',
                'contacts',
            ]
        }),
        ('In-house Info', {
            'fields': [
                'project_manager',
                'engineers',
                'installers',
            ]
        }),
        ('Options', {
            'fields': [
                'modeling',
                'model_upload',
            ]
        }),
    ]
    autocomplete_fields = [
        'addresses',
        'contractors',
        'architects',
        'eors',
        'customers',
        'contacts',
        'engineers',
        'installers',
    ]


stair_fieldsets = [
    (None, {'fields': [(
        'submittal',
        'name',
        'alt_name',
        'grid',
        'levels',
        'stn',
        'stn_note',
        'finish',
        'finish_custom',
    ),
    ]}),
    (None, {'fields': [
        'flights',
        ('stringer_type', 'stringer_custom'),
        ('stair_con', 'stair_con_custom', 'stair_qt',),
    ]}),
    (None, {'fields': [
        'landings',
        ('land_con', 'land_con_custom', 'land_qt',),
    ]}),
    (None, {'fields': [
        ('rail_type', 'rail_type_custom'),
    ]}),
    (None, {'fields': [
        ('stair_fabs', 'land_fabs', 'rail_fabs'),
        ('embeds_shop', 'embeds_na',),
        ('stairs_shop', 'lands_shop', 'rails_shop'),
    ]}),
    (None, {'fields': [
        ('on_hold', 'on_hold_note',),
    ]}),
]


class StairInline(admin.StackedInline):
    model = Stair
    extra = 0
    show_change_link = True
    fieldsets = stair_fieldsets


@admin.register(Submittal)
class SubmittalAdmin(SaveViewMixin, admin.ModelAdmin):
    ordering = ['job__number', 'number']
    inlines = [StairInline]
    search_fields = ['job__number', 'job__name']
    autocomplete_fields = ['job']


@admin.register(Stair)
class StairAdmin(SaveViewMixin, admin.ModelAdmin):
    ordering = ['submittal__job__number', 'submittal__number', 'name']
    autocomplete_fields = ['submittal']
    search_fields = ['submittal__job__number', 'submittal__job__name']
    # inlines = [StairRevInline]
    fieldsets = stair_fieldsets
    list_display = [
        'submittal',
        'name',
        'levels',
        'in_design',
        'in_fab',
        'at_engineering',
        'at_submittal',
        'on_hold',
    ]
    list_display_links = ['name']
    actions = ['put_on_hold', 'remove_hold']

    @admin.action(description='Put selected stairs on hold')
    def put_on_hold(self, _request, queryset):
        queryset.update(on_hold=True)

    @admin.action(description='Remove hold from selected stairs')
    def remove_hold(self, _request, queryset):
        queryset.update(on_hold=False)


class Named(SaveViewMixin, admin.ModelAdmin):
    search_fields = ['name']


@admin.register(Contact)
class ContactAdmin(Named):
    list_display = ['name', 'email', 'company']
    ordering = ['company', 'name']
    autocomplete_fields = ['company']


admin.site.register(Company, Named)
admin.site.register(ProjectManager, Named)


@admin.register(Address)
class AddressAdmin(SaveViewMixin, admin.ModelAdmin):
    list_display = [
        'name',
        'street_address',
        'city',
        'state',
        'zip_code',
    ]
    search_fields = [
        'name',
        'street_address',
        'city',
        'state',
        'zip_code',
    ]
    list_filter = ['city', 'state']


class TransmittalOverrideInline(admin.TabularInline):
    model = TransmittalOverride
    extra = 0
    fieldsets = [
        (None, {'fields': [
            ('stairs',),
            ('response',),
        ]})]
    autocomplete_fields = ['stairs']


@admin.register(Transmittal)
class TransmittalAdmin(SaveViewMixin, admin.ModelAdmin):
    list_display = [
        'job',
        'description',
        'status',
        'to_engr',
        'rec_from_engr',
        'submitted',
        'rec_from_cust',
        'response',
    ]
    autocomplete_fields = [
        'job',
        'stairs',
    ]
    inlines = [TransmittalOverrideInline]


@admin.register(Task)
class TaskAdmin(SaveViewMixin, admin.ModelAdmin):
    list_display = [
        'done',
        'job',
        'description',
        'detailer',
        'stair_names',
        'stage',
        'due',
    ]
    list_display_links = ['job']
    list_filter = [
        'stage',
        'detailer',
    ]
    autocomplete_fields = [
        'job',
        'stairs',
    ]
    search_fields = [
        'job__number',
        'job__name',
    ]
    actions = ['complete']

    @admin.action(description='Mark selected tasks done')
    def complete(self, _request, queryset):
        queryset.update(done=True)

