import json
import os
import re
import subprocess
import tempfile
import sys
from datetime import datetime
from functools import wraps
from glob import glob
from threading import Thread
from time import sleep
from typing import Any, Callable, Optional, Union
from django.conf import settings
from django.core.cache import cache
from django.utils import timezone
from .models import Job, DWGFile
from .utils import first_key

# types

ReplacementList = list[tuple[str, str]]
StrDict = dict[str, Any]
JSONDWG = StrDict
AcObject = StrDict
AcEntity = StrDict
AcXDATA = StrDict
AcSpecialBlock = AcXDATA
SpecialBlocksDict = dict[str, list[AcSpecialBlock]]
TableList = list[list[str]]
TableDict = dict[str, Union[str, TableList]]
Takeoff = dict[str, Union[Optional[SpecialBlocksDict], list[TableDict]]]
DWGTree = dict[str, Union['DWGTree', DWGFile]]

#TODO:
# Point = list[real]
# AcSpecialBlock = dict[str, Union[str, 

# util

def multi_sub(string: str, repls: ReplacementList) -> str:
    'Perform all replacements in REPLS on STRING'
    for frm,to in repls:
        string = re.sub(frm, to, string)
    return string

JSON_REPLS = [
    (r'"VBAProject":\s*\{\s*\{\s*\},', ''),
    (r'-?nan', 'null'),
]

def read_json_file(filename: str) -> str:
    'Read JSON file, making necessary substitutions in JSON_REPLS'
    with open(filename, 'r', errors='ignore') as json:
        return multi_sub(json.read(), JSON_REPLS)

# reverse function composition
# functions called in order specified
def pipeline(*functions):
    def inner(arg):
        for f in functions:
            arg = f(arg)
        return arg
    return inner

def get_obj(json_data: JSONDWG, handle: list[int]) -> Optional[AcEntity]:
    'Find and return a single object by handle from a JSON DWG, or None if not found'
    # not sure what's at play with handle formatting
    # but this seems to work...
    handle = [0, 3, handle[2]]
    for obj in json_data['OBJECTS']:
        if obj['handle'] == handle:
            return obj
    return None

# common predicate

def is_entity(obj: AcObject) -> bool:
    return 'entity' in obj.keys()

# special block handling

def keyfilter(func):
    @wraps(func)
    def wrapper(x):
        if x is None:
            return ''
        return func(x)
    return wrapper

@keyfilter
def round2(n):
    return round(n, 2)

@keyfilter
def bool01(x):
    if x=='1':
        return '✔️'
    return ''

NAME_FIELDS = ['assembly', 'style', 'guard_ass']

SummaryKeyConfig = Union[tuple[str, str], tuple[str, str, Callable]]
AppIDConfig = dict[tuple[str, ...], list[SummaryKeyConfig]]
APPIDS: AppIDConfig = {
    ('Stair', '3D_Stair'): [
        ('rise_hgt', 'Rise Ht', round2),
        ('rise_qty', 'Risers'),
        ('width', 'Width', round2),
        ('tread_style', 'Tread Style'),
        ('bot_conn', 'Bot Connection'),
        ('top_conn', 'Top Connection'),
    ],
    ('Land', '3D_Land'): [
        ('plt_wid', 'Width', round2),
        ('plt_dpt', 'Depth', round2),
        ('tread_style', 'Tread Style'),
    ],
    ('Guard', '3D_Guard'): [
        ('guard_qty', 'Qty'),
        ('guard_lng', 'Length', round2),
        ('guard_hgt', 'Height', round2),
        ('guard_hand', 'Handrail', bool01),
        ('guard_kick', 'Kick Plate', bool01),
    ],
    ('Embed', '3D_Embed'): [
        ('t_lng', 'Length', round2),
    ],
}
APPID_SUMMARY_KEYS = {
    appid: cols+[('comment','Comment')]
    for appids, cols in APPIDS.items()
    for appid in appids
}
DESIRED_APPIDS = APPID_SUMMARY_KEYS.keys()

def is_special_block(obj: AcObject) -> bool:
    if not is_entity(obj):
        return False
    if obj['entity'] != 'INSERT':
        return False
    if 'eed' not in obj.keys():
        return False
    if any([xd for xd in obj['eed'] if xd['value'] == 'app_id']):
        return True
    return False

def is_desired_special_block(obj: AcObject) -> bool:
    return is_special_block(obj) and format_xdata(obj)['app_id'] in DESIRED_APPIDS

def format_spec_block_factory(json_data: JSONDWG) -> Callable[[AcEntity], StrDict]:
    '''Return a function which accepts an entity and returns a formatted dict of its properties

    json_data is used by the resulting function to look up related entities'''
    def format_spec_block(entity: AcEntity) -> StrDict:
        'Return entity properties we want to extract as a dictionary'
        ins = entity['ins_pt']
        xdata = format_xdata(entity)
        layer_ent = get_obj(json_data, entity['layer'])
        if layer_ent is None:
            layer = ''
        else:
            layer = layer_ent['name']	# type: ignore
        if xdata['app_id'][0:2] == '3D':
            sort_key = ins[2]
        else:
            sort_key = ins[1]
        name = first_key(xdata, NAME_FIELDS) or 'Unknown'
        summary_keys = APPID_SUMMARY_KEYS[xdata['app_id']]
        return {
            'ins': ins,
            'xdata': xdata,
            'sort_key': sort_key,
            'layer': layer,
            'name': name,
            'summary_vals': [
                k[2](xdata.get(k[0])) if len(k) > 2 else xdata.get(k[0], '') # type: ignore
                for k in summary_keys
            ],
        }
    return format_spec_block

def format_xdata(entity: AcEntity) -> AcXDATA:
    'Return XDATA from ENTITY as a dictionary'
    eed = entity['eed']
    return {k['value']: v['value'] for k,v in list(zip(eed[::2], eed[1::2]))}

def get_special_blocks(json_data: JSONDWG) -> Optional[SpecialBlocksDict]:
    'Return special blocks in JSON_DATA as a list of XDATA dictionaries'
    blocks = sorted(
        map(
            format_spec_block_factory(json_data),
            filter(is_desired_special_block, json_data['OBJECTS'])
        ),
        key=lambda b: b['sort_key'],
        reverse=True,
    )
    appids: SpecialBlocksDict = {appid: [] for appid in DESIRED_APPIDS}
    for block in blocks:
        appids[block['xdata']['app_id']].append(block)
    appids = {k: v for k, v in appids.items() if v != []}
    return appids or None

# extract tables

def is_table_block(obj: AcObject) -> bool:
    return 'object' in obj.keys() and obj['object'] == 'BLOCK_HEADER' and obj['name']=='*T'

def get_table_blocks(json_data: JSONDWG) -> list[AcEntity]:
    'Return table blocks as a list'
    return list(filter(is_table_block, json_data['OBJECTS']))

def get_tables(json_data: JSONDWG) -> list[TableDict]:
    'Return tables as a list of tables, each of which is a list of lists of strings'
    blocks = get_table_blocks(json_data)
    tables = []
    for block in blocks:
        ents = []
        for handle in block['entities']:
            ent = get_obj(json_data, handle)
            if ent and ent['entity'] == 'MTEXT':
                ents.append(ent)
        if ents != []:
            tables.append(ents)
    return list(map(normalize_table, tables))

def calculate_rows_and_cols(table_list: list[AcEntity]) -> TableList:
    '''Assign row and column indices to children of table

    Each child of a table block is an MTEXT entity. We can (mostly)
    figure out what row and column each exists in by its relative
    insert point'''
    col_set = set()
    row_set = set()
    for mtext in table_list:
        ins = mtext['ins_pt']
        ins[0] = round(ins[0], 3)
        ins[1] = round(ins[1], 3)
        col_set.add(ins[0])
        row_set.add(ins[1])
    cols = sorted(col_set)
    rows = sorted(row_set, reverse=True)
    tbl = [['' for _ in cols] for _ in rows]
    for mtext in table_list:
        ins = mtext['ins_pt']
        col = cols.index(ins[0])
        row = rows.index(ins[1])
        tbl[row][col] = mtext['text']
    return tbl

def remove_phantom_column(tbl: TableList) -> TableList:
    '''Account for phantom columns in table

    When using insert points to determine row and column membership
    (see calculate_rows_and_cols()), a merged+centered title row can
    indicate the presence of a column that does not exist in other
    rows. We strip those out in this step.'''
    phantom_col_idx = None
    for i,col in enumerate(list(zip(*tbl[1:]))):
        if all([x=='' for x in col]):
            phantom_col_idx = i
            break
    if phantom_col_idx:
        # if we have a phantom column, it's because of a title row
        # which means every col in that row should be empty except for the phantom
        tbl[0][0] = tbl[0][phantom_col_idx]
        for row in tbl:
            del row[phantom_col_idx]
    return tbl

def table_list_to_dict(tbl: TableList) -> TableDict:
    'Returns the table in dictionary format'
    # the first row is a title row if it has exactly one entry
    title_row = [x for x in tbl[0] if x!='']
    if len(title_row) == 1:
        return {
            'title': title_row[0],
            'body': tbl[1:],
        }
    else:
        return {'body': tbl}

normalize_table = pipeline(
    calculate_rows_and_cols,
    remove_phantom_column,
    table_list_to_dict,
)

# main

def extract_data(json_data: JSONDWG) -> Takeoff:
    'Return the information we want as a serialized JSON object'
    return {
        'blocks': get_special_blocks(json_data),
        'tables': get_tables(json_data),
    }

def take_off_dwg(filename: str) -> Takeoff:
    'Use `dwgread` to produce JSON output of a DWG and extract our data'
    base = os.path.basename(filename)
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpfile = os.path.join(tmpdir, base)[:-3] + 'json'
        subprocess.run(['dwgread', filename, '-o', tmpfile], capture_output=True)
        return extract_data(json.loads(read_json_file(tmpfile)))

def scan_job(number: int, force: bool=False) -> None:
    '''Scan a job folder and update the database

    Files older than our cache are skipped. Pass force=True to
    override and re-scan all files anyway.

    To be nice to the filesystem and network, a 0.5s delay is inserted
    between file scans (but not file stats for mtime checking).
    '''
    job = Job.objects.get(number=number)
    jobdir = f'Job{number}'
    files = glob(
        os.path.join(
            settings.DETAIL_PATH,
            'Drawings/Current Drawings',
            jobdir,
            '**/*.dwg',
        ),
        recursive=True,
    )
    for f in files:
        try:
            dwgfile = DWGFile.objects.get(job=job, path=f)
            new = False
        except DWGFile.DoesNotExist:
            dwgfile = DWGFile(job=job, path=f)
            new = True
        mtime = timezone.make_aware(datetime.fromtimestamp(os.path.getmtime(f)))
        if not force and not new and mtime <= dwgfile.updated:
            continue
        try:
            dwgfile.takeoff = take_off_dwg(f)
        except Exception as e:
            print(f'Error attempting to take off DWG: {f}\n{e}')
            continue
        dwgfile.updated = timezone.now()
        dwgfile.save()
        sleep(0.5)

class ScanJobMixin:
    'Mixin for view that initiates a job scan'
    def get(self, request, *args, **kwargs):
        number = kwargs['pk']
        if cache.add(f'scan{number}', True, 300):
            Thread(target=scan_job, args=[number]).start()
        return super().get(request, *args, **kwargs)

def dwg_tree(dwgs_qset) -> DWGTree:
    # mutates tree dict
    def _recurse(dic: dict, chain: list[str]):
        if len(chain) == 0:
            return
        if len(chain) == 1:
            dic[chain[0]] = dwg
            return
        key, *new_chain = chain
        if key not in dic:
            dic[key] = {}
        _recurse(dic[key], new_chain)
        return

    tree: DWGTree = {}
    for dwg in dwgs_qset:
        _recurse(tree, dwg.path[41:].split('/'))
    return tree

