import json
import os
from datetime import datetime, timedelta
from django.conf import settings
from django.contrib.admin.decorators import display
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_save
from django.urls import reverse
from django.utils.functional import cached_property
from functools import reduce
from localflavor.us.us_states import STATE_CHOICES
from pathlib import Path
import re
from phonenumber_field.modelfields import PhoneNumberField
from .utils import render_tooltip


class Address(models.Model):
    name = models.CharField(max_length=100, default='Jobsite')
    street_address = models.CharField(max_length=100)
    city = models.CharField(max_length=100)
    state = models.CharField(max_length=2, choices=STATE_CHOICES)
    zip_code = models.CharField(max_length=5)

    def location(self):
        return f'{self.city}, {self.state}'

    def __str__(self):
        return '%s\n%s\n%s %s' % (
            self.name,
            self.street_address,
            self.location(),
            self.zip_code,
        )

    def query_string(self):
        return '?q=%s+%s+%s+%s&t=h' % (
            self.street_address,
            self.city,
            self.state,
            self.zip_code,
            )

    class Meta:
        verbose_name_plural = 'addresses'
        unique_together = [[
            'street_address',
            'city',
            'state',
            'zip_code',
        ]]
        ordering = ['state', 'zip_code', 'street_address']


class Company(models.Model):
    '''A company we do business with'''
    name = models.CharField(max_length=100)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name_plural = 'companies'
        ordering = ['name']


class Person(models.Model):
    '''Abstract class representing any person'''
    name = models.CharField(max_length=100)
    email = models.EmailField(blank=True)

    def __str__(self):
        return self.name

    class Meta:
        abstract = True
        ordering = ['name']


class Contact(Person):
    '''A contact at a client company'''
    company = models.ForeignKey(Company, on_delete=models.PROTECT)
    phone = PhoneNumberField(blank=True, null=True)
    job_title = models.CharField(max_length=200, blank=True, null=True)

    def details(self):
        return f'{self.email}\n{self.phone or ""}'

    def get_absolute_url(self):
        return reverse('jobs:company_detail', args=[self.company.id])


class Detailer(User):
    '''A PSC detailer'''
    class Meta:
        proxy = True
        ordering = ['first_name']

    def name(self):
        return f'{self.first_name} {self.last_name}'

    def sub_jobs(self):
        return [
            Job.objects.get(pk=pk)
            for pk in
            set(self.task_set.filter(stage='SUB').values_list('job', flat=True))
        ]

    def fab_jobs(self):
        return [
            Job.objects.get(pk=pk)
            for pk in
            set(self.task_set.filter(stage='FAB').values_list('job', flat=True))
        ]

    def tasks(self):
        return Task.objects.filter(detailer__id=self.id)

    def open_tasks(self):
        return Task.objects.filter(detailer__id=self.id, done=False)

    def __str__(self):
        return self.name()


class ProjectManager(Person):
    '''A PSC project manager'''

    def get_absolute_url(self):
        return reverse('jobs:pm_detail', args=[self.id])


class CurrentJobsManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(archived=False)


class ArchivedJobsManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(archived=True)


class Job(models.Model):
    objects = models.Manager()
    current = CurrentJobsManager()
    archive = ArchivedJobsManager()
    number = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=100)
    addresses = models.ManyToManyField(
        Address,
        blank=True,
    )
    contractors = models.ManyToManyField(
        Company,
        related_name='contractor_jobs',
        blank=True,
    )
    architects = models.ManyToManyField(
        Company,
        related_name='architect_jobs',
        blank=True,
    )
    eors = models.ManyToManyField(
        Company,
        related_name='eor_jobs',
        verbose_name='engineer of record',
        blank=True,
    )
    customers = models.ManyToManyField(
        Company,
        related_name='customer_jobs',
        blank=True,
    )
    contacts = models.ManyToManyField(Contact, blank=True)
    project_manager = models.ForeignKey(
        ProjectManager,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
    )
    engineers = models.ManyToManyField(
        Company,
        related_name='engineer_jobs',
        blank=True,
    )
    installers = models.ManyToManyField(
        Company,
        related_name='installer_jobs',
        blank=True,
    )
    award_date = models.DateField()

    modeling = models.BooleanField('3D modeling included', default=False)
    model_upload = models.CharField(
        '3D model upload link',
        max_length=200,
        blank=True,
    )
    archived = models.BooleanField(default=False)

    def get_absolute_url(self):
        return reverse('jobs:job_detail', args=[self.number])

    def get_correspondence(self):
        try:
            jpath = list(Path(settings.PM_PATH).glob(f'{self.number}*'))[0]
        except: # noqa
            return None
        corr = []

        def corr_item(ctype, p):
            text = p.name.replace('_', ' ')
            match = re.search(r'([0-9]{6}) (.+)', text)
            if 'To' in ctype:
                color = 'green'
            elif 'From' in ctype:
                color = 'darkred'
            else:
                color = 'inherit'
            if match is None:
                date = ''
                date_raw = ''
                date_str = ''
                desc = p.name
            else:
                date_raw = match[1]
                date = datetime.strptime(match[1], '%y%m%d')
                date_str = date.strftime('%Y-%m-%d')
                desc = match[2]
            return {
                'type': ctype,
                'color': color,
                'date_raw': date_raw,
                'date_str': date_str,
                'date': date,
                'desc': desc,
                'link': f'file:///{str(p).replace(settings.PM_PATH, "P:")}',
            }

        if jpath.exists():
            for ctype in (jpath / '01 Correspondence').iterdir():
                ctypename = ctype.name[3:]
                corr.extend([corr_item(ctypename, p) for p in ctype.iterdir()])
            return sorted(corr, key=lambda d: d['date_raw'], reverse=True)
        else:
            return []

    def dates(self):
        return self.jobdate_set.order_by('date')

    def sub_tasks(self):
        return self.task_set.filter(stage='SUB')

    def fab_tasks(self):
        return self.task_set.filter(stage='FAB')

    def sub_detailers(self):
        detailers = self.task_set.filter(stage='SUB').values_list('detailer__id', flat=True)
        return [Detailer.objects.get(pk=d) for d in set(detailers)]

    def fab_detailers(self):
        detailers = self.task_set.filter(stage='FAB').values_list('detailer__id', flat=True)
        return [Detailer.objects.get(pk=d) for d in set(detailers)]

    def total_flights(self):
        subs = self.submittal_set.all()
        return reduce(lambda x, y: x+y, [s.total_flights() for s in subs])

    def total_landings(self):
        subs = self.submittal_set.all()
        return reduce(lambda x, y: x+y, [s.total_landings() for s in subs])

    def get_stairs(self):
        return Stair.objects.filter(submittal__job__number=self.number)

    def get_stairs_qstring(self):
        return ','.join([str(s.pk) for s in self.get_stairs()])

    def stairs_in_design(self):
        return list(filter(lambda s: s.in_design(), self.get_stairs()))

    def stairs_in_fab(self):
        return [s.unique_name() for s in self.get_stairs() if s.in_fab()]

    def stairs_at_engineering(self):
        return [s.unique_name() for s in self.get_stairs() if s.at_engineering()]

    def stairs_at_submittal(self):
        return [s.unique_name() for s in self.get_stairs() if s.at_submittal()]

    def __str__(self):
        return f"{self.number} - {self.name}"

    def post_save_handler(sender, **kwargs):
        job = kwargs['instance']
        if job.submittal_set.count() == 0:
            Submittal.objects.create(job=job, number=1)

    def valid_dwgs(self, request):
        params = request.GET.dict()
        q = self.drawings.filter(updated__isnull=False)
        if params.get('include_empty') != 'on':
            q = q.exclude(takeoff__blocks=None, takeoff__tables=[])
        appids = [
            'Stair',
            '3D_Stair',
            'Land',
            '3D_Land',
            'Guard',
            '3D_Guard',
            'Embed',
            '3D_Embed',
        ]
        filter_appids = appids & params.keys()
        if filter_appids:
            return q.filter(takeoff__blocks__has_any_keys=filter_appids)
        return q

    def is_in_design(self):
        return any([xmtl.status() == 'In Design' for xmtl in self.transmittal_set.all()])

    def is_in_fab(self):
        return any([stair.in_fab() for sub in self.submittal_set.all() for stair in sub.stair_set.all()])

    def is_at_engineering(self):
        return any([xmtl.status() == 'Engineering' for xmtl in self.transmittal_set.all()])

    def is_at_submittal(self):
        return any([xmtl.status() == 'Submitted' for xmtl in self.transmittal_set.all()])

    def status_color(self):
        if self.is_in_design():
            return 'blue'
        if self.is_in_fab():
            return 'red'
        if self.is_at_engineering():
            return 'green'
        if self.is_at_submittal():
            return 'orange'
        return 'initial'

    class Meta:
        ordering = ['-number']


post_save.connect(Job.post_save_handler, sender=Job)


class JobDate(models.Model):
    job = models.ForeignKey(
        Job,
        on_delete=models.CASCADE,
    )
    date = models.DateField()
    description = models.CharField(max_length=200)
    note = models.CharField(max_length=200, blank=True, null=True)


class JobNote(models.Model):
    job = models.ForeignKey(Job, on_delete=models.CASCADE)
    date = models.DateTimeField(auto_now=True)
    note = models.TextField(max_length=1000)

    def __str__(self):
        return self.note


class Construction(models.Model):
    job = models.ForeignKey(Job, on_delete=models.CASCADE)
    STYLE_CHOICES = [
        ('concrete', 'Concrete'),
        ('steel', 'Steel'),
        ('wood', 'Wood'),
        ('custom', 'Custom'),
    ]
    style = models.CharField(max_length=20, choices=STYLE_CHOICES)
    style_custom = models.CharField(max_length=100, blank=True, null=True)
    scope = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        return '%s' % self.style_custom if self.style == 'CU' else self.get_style_display()


class Submittal(models.Model):
    class Meta:
        ordering = ['job', 'number']

    job = models.ForeignKey(Job, on_delete=models.CASCADE)
    number = models.IntegerField(default=1)
    name = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        return f"{self.job} sub {self.number}"

    def label(self):
        if self.name:
            return f'Sub {self.number} - {self.name}'
        else:
            return f'Sub {self.number}'

    def total_flights(self):
        return sum([
            s.flights for s in self.stair_set.filter(on_hold=False).exclude(stair_con='BO')
        ])

    def total_landings(self):
        return sum([
            s.landings for s in self.stair_set.filter(on_hold=False).exclude(land_con='BO')
        ])

    def get_absolute_url(self):
        return reverse('jobs:job_detail', args=[self.job.number])

    def get_stairs_qstring(self):
        return ','.join([
            str(x)
            for x in self.stair_set.filter(on_hold=False).values_list('pk', flat=True)
        ])

    def calendar(self):
        return [
            {
                'name': stair.unique_name,
                'xmtls': [
                    {
                        'pk': xmtl.pk,
                        'description': xmtl.description,
                        'status': xmtl.status,
                        'to_engr': xmtl.to_engr,
                        'due_from_engr': xmtl.due_from_engr_display,
                        'rec_from_engr': xmtl.rec_from_engr,
                        'submitted': xmtl.submitted,
                        'rec_from_cust': xmtl.rec_from_cust_display(),
                        'response': xmtl.get_response(stair),
                    }
                    for xmtl in stair.transmittal_set.order_by('-pk')
                ]
            }
            for stair in self.stair_set.filter(on_hold=False)
        ]


class Stair(models.Model):
    class Meta:
        ordering = ['submittal__number', 'sub_order', 'pk']

    submittal = models.ForeignKey(Submittal, on_delete=models.CASCADE)
    sub_order = models.IntegerField(default=0)
    name = models.CharField(max_length=100)
    alt_name = models.CharField(max_length=100, blank=True, null=True)
    grid = models.CharField(max_length=100, blank=True)
    levels = models.CharField(max_length=100, blank=True)
    flights = models.IntegerField(default=0)
    STRINGER_CHOICES = [
        ('C', 'C Channel'),
        ('MC', 'MC Channel'),
        ('PL', 'Plate'),
        ('CU', 'Custom'),
    ]
    stringer_type = models.CharField(
        max_length=2,
        choices=STRINGER_CHOICES,
        default='PL',
    )
    stringer_custom = models.CharField(max_length=50, blank=True)

    def stringer_display(self):
        if self.flights == 0:
            return ''
        title = None
        if self.stringer_type == 'CU':
            stype = self.stringer_custom
        else:
            stype = self.get_stringer_type_display()
            title = self.stringer_custom
        return render_tooltip(stype, title)

    CONSTRUCTION_CHOICES = [
        ('PF', 'Pan-filled'),
        ('CP', 'Checker'),
        ('PC', 'Precast'),
        ('GT', 'Grating'),
        ('CU', 'Custom'),
        ('BO', 'By other'),
    ]
    stair_con = models.CharField(
        'stair construction',
        max_length=2,
        choices=CONSTRUCTION_CHOICES,
        default='PF',
    )
    stair_con_custom = models.CharField(
        'custom stair construction',
        max_length=50,
        blank=True
    )

    def stair_con_display(self):
        if self.flights == 0:
            return ''
        title = None
        if self.stair_con == 'CU':
            con = self.stair_con_custom
        else:
            con = self.get_stair_con_display()
            title = self.stair_con_custom
        if self.stair_qt:
            con = con + ' [QT]'
        return render_tooltip(con, title)

    stair_qt = models.BooleanField('stair quiet tread', default=False)
    landings = models.IntegerField(default=0)
    land_con = models.CharField(
        'landing construction',
        max_length=2,
        choices=CONSTRUCTION_CHOICES,
        default='PF',
    )
    land_con_custom = models.CharField(
        'custom landing construction',
        max_length=50,
        blank=True
    )

    def land_con_display(self):
        if self.landings == 0:
            return ''
        title = None
        if self.land_con == 'CU':
            con = self.land_con_custom
        else:
            con = self.get_land_con_display()
            title = self.land_con_custom
        if self.land_qt:
            con = con + ' [QT]'
        return render_tooltip(con, title)

    land_qt = models.BooleanField('landing quiet tread', default=False)
    RAIL_CHOICES = [
        ('1', '100 Pipe'),
        ('2', '200 Pipe'),
        ('3', '300 Rec Tube'),
        ('4', '400 Picket'),
        ('5', '500 Mesh'),
        ('6', '600 Cable'),
        ('7', '700 Rod'),
        ('8', '800 Glass'),
        ('W', 'Wall rail only'),
        ('C', 'Custom'),
        ('B', 'By other'),
    ]
    rail_type = models.CharField(
        max_length=1,
        choices=RAIL_CHOICES,
        default='2',
    )
    rail_type_custom = models.CharField(max_length=50, blank=True)

    def rail_type_display(self):
        title = None
        if self.rail_type == 'C':
            rtype = self.rail_type_custom
        else:
            rtype = self.get_rail_type_display()
            title = self.rail_type_custom
        return render_tooltip(rtype, title)

    stn = models.CharField('STN', max_length=50, blank=True)
    stn_note = models.CharField('STN note', max_length=200, blank=True, null=True)

    FINISH_CHOICES = [
        ('rodda red oxide', 'Rodda Red Oxide'),
        ('rodda black', 'Rodda Black'),
        ('rodda grey', 'Rodda Grey'),
        ('rodda high solids red oxide', 'Rodda High Solids Red Oxide'),
        ('rodda high solids black', 'Rodda High Solids Black'),
        ('rodda high solids grey', 'Rodda High Solids Grey'),
        ('galvanized', 'Galvanized'),
        ('custom', 'Custom'),
    ]

    finish = models.CharField(
        choices=FINISH_CHOICES,
        max_length=50,
        default='rodda red oxide',
    )
    finish_custom = models.CharField(max_length=50, blank=True)

    def finish_display(self):
        if self.finish == 'custom':
            return self.finish_custom
        return render_tooltip(self.get_finish_display(), self.finish_custom)

    stair_fabs = models.DateField(
        'stair fabs done',
        help_text='N/A when 0 flights or flights B/O',
        blank=True,
        null=True,
    )
    land_fabs = models.DateField(
        'land fabs done',
        help_text='N/A when 0 lands or lands B/O',
        blank=True,
        null=True,
    )
    rail_fabs = models.DateField(
        'rail fabs done',
        help_text='N/A when rails B/O',
        blank=True,
        null=True,
    )
    stairs_shop = models.DateField(
        'stairs to shop',
        help_text='N/A when 0 flights or flights B/O',
        blank=True,
        null=True,
    )
    lands_shop = models.DateField(
        'lands to shop',
        help_text='N/A when 0 lands or lands B/O',
        blank=True,
        null=True,
    )
    rails_shop = models.DateField(
        'rails to shop',
        help_text='N/A when rails B/O',
        blank=True,
        null=True,
    )

    embeds_shop = models.DateField('embeds to shop', blank=True, null=True)
    embeds_na = models.BooleanField(verbose_name='Embeds not applicable', default=False)

    def stairs_na(self):
        return self.flights == 0 or self.stair_con == 'BO'

    def landings_na(self):
        return self.landings == 0 or self.land_con == 'BO'

    def rails_na(self):
        return self.rail_type == 'B'

    on_hold = models.BooleanField(default=False)
    on_hold_note = models.CharField(max_length=200, blank=True, null=True)

    def unique_name(self):
        if self.submittal.job.get_stairs().filter(name=self.name).count() > 1:
            return f'{self.name} ({self.levels})'
        return self.name

    def get_absolute_url(self):
        return reverse('jobs:job_detail', args=[self.submittal.job.number])

    # statuses
    def _block_on_hold(self, stage=None, status=None):
        if self.on_hold:
            return False
        else:
            if stage is not None:
                return any([not t.done for t in self.task_set.filter(stage=stage)])
            else:
                return any([xmtl.status() == status for xmtl in self.transmittal_set.all()])

    def has_open_calendar_slots(self):
        if not self.stairs_na() and (not self.stair_fabs or not self.stairs_shop):
            return True
        if not self.landings_na() and (not self.land_fabs or not self.lands_shop):
            return True
        if not self.rails_na() and (not self.rail_fabs or not self.rails_shop):
            return True
        if not self.embeds_na and not self.embeds_shop:
            return True
        return False

    @display(boolean=True)
    def in_design(self):
        return not self.on_hold and self._block_on_hold(stage='SUB')

    @display(boolean=True)
    def in_fab(self):
        if self.on_hold:
            return False
        my_xmtls = Transmittal.objects.filter(stairs__in=[self])
        fab_available = any([xmtl.response in ['NC', 'AN'] or xmtl.for_record for xmtl in my_xmtls])
        if not fab_available:
            return False
        return self.has_open_calendar_slots()

    @display(boolean=True)
    def at_engineering(self):
        return self._block_on_hold(status='Engineering')

    @display(boolean=True)
    def at_submittal(self):
        return self._block_on_hold(status='Submitted')

    def clean(self):
        other_stairs = [
            (s.name, s.levels) for s in self.submittal.job.get_stairs().exclude(id=self.id)
        ]
        if (self.name, self.levels) in other_stairs:
            raise ValidationError('Name and levels must be unique in a job')

    def __str__(self):
        job = self.submittal.job.number
        sub = self.submittal.number
        return f"{job} sub {sub} stair {self.unique_name()}"


class StairNamesMixin:
    @cached_property
    def stair_names(self):
        stairs_qset = self.stairs.filter(on_hold=False)
        stairs = list(stairs_qset)
        return ', '.join([s.unique_name() for s in stairs])


class Transmittal(models.Model, StairNamesMixin):
    class Meta:
        ordering = ['job', 'description', 'to_engr', 'id']

    def get_absolute_url(self):
        return reverse('jobs:job_detail', args=[self.job.number])

    job = models.ForeignKey(Job, on_delete=models.CASCADE)
    stairs = models.ManyToManyField(Stair)
    description = models.CharField(max_length=200)
    to_engr = models.DateField('sent to engineer', blank=True, null=True)
    due_from_engr = models.DateField('due from engineer', blank=True, null=True)
    rec_from_engr = models.DateField(
        'received from engineer',
        blank=True,
        null=True,
    )
    submitted = models.DateField('submitted', blank=True, null=True)
    rec_from_cust = models.DateField(
        'received from customer',
        blank=True,
        null=True,
    )
    RESPONSE_CHOICES = [
        ('NC', 'Approved'),
        ('AN', 'Approved as noted'),
        ('RR', 'Revise and resubmit'),
        ('U', 'Update'),
    ]
    response = models.CharField(
        max_length=2,
        choices=RESPONSE_CHOICES,
        blank=True,
    )
    for_record = models.BooleanField(default=False)

    def due_from_engr_display(self):
        if self.to_engr is None:
            return ''
        if self.due_from_engr is not None:
            return self.due_from_engr
        else:
            return self.to_engr + timedelta(days=10)

    def status(self):
        if self.rec_from_cust:
            return 'Returned'
        elif self.submitted:
            if self.for_record:
                return 'For record'
            else:
                return 'Submitted'
        elif self.rec_from_engr:
            return 'Ready to Submit'
        elif self.to_engr:
            return 'Engineering'
        else:
            return 'In Design'

    def effective_date(self):
        if self.status() == 'Returned':
            return self.rec_from_cust
        if self.status() == 'Submitted':
            return self.submitted
        if self.status() == 'Ready to Submit':
            return self.rec_from_engr
        if self.status() == 'Engineering':
            return self.to_engr

    def get_response(self):
        response = self.get_response_display()
        if response:
            return response
        if self.for_record:
            return 'N/A'
        return ''

    def rec_from_cust_display(self):
        if self.rec_from_cust is not None:
            return self.rec_from_cust.strftime('%m/%d/%Y')
        if self.for_record:
            return 'N/A'

    def __str__(self):
        return f'{self.job.name} transmittal (#{self.pk})'


class TransmittalOverride(models.Model):
    transmittal = models.ForeignKey(Transmittal, on_delete=models.CASCADE)
    stairs = models.ManyToManyField(Stair)
    response = models.CharField(
        max_length=2,
        choices=Transmittal.RESPONSE_CHOICES,
    )


class Task(models.Model, StairNamesMixin):
    class Meta:
        ordering = ['done', 'due', '-id']

    def get_absolute_url(self):
        return reverse('jobs:job_detail', args=[self.job.number])

    detailer = models.ForeignKey(Detailer, on_delete=models.PROTECT)
    job = models.ForeignKey(Job, on_delete=models.CASCADE)
    stairs = models.ManyToManyField(Stair)
    description = models.CharField(max_length=100, blank=True, null=True)
    stage = models.CharField(
        max_length=3,
        choices=[('SUB', 'Submittal'), ('FAB', 'Fab')],
    )
    due = models.DateField(blank=True, null=True)
    done = models.BooleanField(default=False)

    def __str__(self):
        return f'[{self.job.number}] {self.stair_names} {self.description or ""}'
