from django.contrib import messages
from django.core.exceptions import BadRequest
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.views import generic
from django.views.decorators.http import require_http_methods
from .models import (
    Job,
    Submittal,
    Detailer,
    ProjectManager,
    Company,
    Task,
    Stair,
    Transmittal,
)
from .forms import StairScopeRowForm, TransmittalForm, make_transmittal_form

def _get_msg_response():
    return HttpResponse(headers={'HX-Trigger': 'getmessages'})

def _form_error_msgs(request, model_name, form):
    response_msg = f'Failed to save new {model_name}'
    for k, v in form.errors.items():
        response_msg += f'<br />{k}: {v}'
    messages.error(request, mark_safe(response_msg))
    return _get_msg_response()

def _delete_msg(request, obj):
    name = str(obj)
    obj.delete()
    messages.success(request, f'{name} deleted')
    return _get_msg_response()


class JobIndexView(generic.ListView):
    template_name = 'jobs/job_index.html'
    context_object_name = 'jobs'

    def get_queryset(self):
        return Job.current.all()


class JobArchiveView(generic.ListView):
    template_name = 'jobs/job_archive.html'
    context_object_name = 'jobs'

    def get_queryset(self):
        return Job.archive.all()

class JobDetailView(generic.DetailView):
    model = Job
    template_name = 'jobs/job_detail.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        job = context['job']
        new_context = dict(
            contractors=job.contractors.all(),
            architects=job.architects.all(),
            eors=job.eors.all(),
            customers=job.customers.all(),
            jobs=Job.objects.filter(archived=False).order_by('number'),
        )
        return context | new_context

class DetailerDetailView(generic.DetailView):
    model = Detailer
    template_name = 'jobs/detailer_detail.html'


class PMDetailView(generic.DetailView):
    model = ProjectManager
    template_name = 'jobs/pm_detail.html'
    context_object_name = 'pm'


class CompanyDetailView(generic.DetailView):
    model = Company
    template_name = 'jobs/company_detail.html'
    context_object_name = 'company'


def complete_task(request, pk):
    user = request.user
    task = Task.objects.get(pk=pk)
    if user.is_authenticated:
        if user.pk == task.detailer.pk:
            task.done = True
            task.save()
    return HttpResponseRedirect(reverse('jobs:job_detail', args=(task.job.number,)))


def job_report(request):
    jobs = Job.objects.exclude(archived=True)
    return render(
        request,
        'jobs/job_report.html',
        {
            'in_design': [
                {
                    'job': job,
                    'tasks': job.sub_tasks().filter(done=False)
                }
                for job in (filter(lambda j: j.stairs_in_design(), jobs))
            ],
            'in_fab': [
                {
                    'job': job,
                    'tasks': job.fab_tasks().filter(done=False)
                }
                for job in (filter(lambda j: j.stairs_in_fab(), jobs))
            ],
            'at_engineering': [
                {
                    'job': job,
                    'xmtls': list(filter(
                        lambda xmtl: xmtl.status() == 'Engineering',
                        Transmittal.objects.filter(job_id=job.number)
                    ))
                }
                for job in (filter(lambda j: j.stairs_at_engineering(), jobs))
            ],
            'at_submittal': [
                {
                    'job': job,
                    'xmtls': list(filter(
                        lambda xmtl: xmtl.status() == 'Submitted',
                        Transmittal.objects.filter(job_id=job.number)
                    ))
                }
                for job in (filter(lambda j: j.stairs_at_submittal(), jobs))
            ],
            'tasks': Task.objects.filter(done=False).order_by('detailer__last_name'),
        },
    )


@require_http_methods(['GET'])
def sub_scope_total(request, pk):
    if 'type' not in request.GET:
        return BadRequest('type parameter is required')
    sl = request.GET['type']
    if sl not in ['flights', 'landings']:
        return BadRequest('bad type parameter')
    sub = Submittal.objects.get(pk=pk)
    if sl == 'flights':
        return HttpResponse(sub.total_flights())
    return HttpResponse(sub.total_landings())

# pk is for sub unless adding, then it is for job
def sub_scope(request, pk):
    if request.method == 'DELETE':
        Submittal.objects.get(pk=pk).delete()
        return HttpResponse()
    if 'add' in request.GET:
        job = Job.objects.get(pk=pk)
        sub = Submittal.objects.create(
            job=job,
            number=job.submittal_set.count() + 1,
        )
    else:
        sub = Submittal.objects.get(pk=pk)
    context = {
        'sub': sub,
        'bodyid': f'scope_sub_{sub.number}'
    }
    return render(request, 'jobs/sub_scope.html', context)

def sub_update_order(request, job_pk):
    subs = list(Job.objects.get(pk=job_pk).submittal_set.all())
    event = request.POST.dict()
    oldidx = int(event['oldidx'])
    newidx = int(event['newidx'])
    src = subs[int(event['source']) - 1]
    src_stairs = list(src.stair_set.all())
    fields = ['sub_order']
    if event['moved'] == 'true':
        dest = subs[int(event['dest']) - 1]
        dest_stairs = list(dest.stair_set.all())
        fields.append('submittal')
    else:
        dest = None
        dest_stairs = src_stairs
    dest_stairs.insert(newidx, src_stairs.pop(oldidx))
    for i, stair in enumerate(src_stairs):
        stair.sub_order = i
    if dest:
        for i, stair in enumerate(dest_stairs):
            stair.submittal = dest
            stair.sub_order = i
    all_stairs = set(src_stairs) | set(dest_stairs)
    Stair.objects.bulk_update(all_stairs, fields)
    return HttpResponse(headers={'HX-Trigger': 'sort_stairs'})

def stair_scope_row(request, pk):
    stair = Stair.objects.get(pk=pk)
    context = {'stair': stair}
    if request.method == 'DELETE':
        return _delete_msg(request, stair)
    if request.method == 'POST':
        form = StairScopeRowForm(request.POST, instance=stair)
        if not form.errors:
            form.save()
            response = render(request, 'jobs/stair_scope_row.html', context)
            response['HX-Trigger'] = 'update_sub_totals'
            return response
        else:
            response_msg = f'Failed to save changes to stair {stair.name}'
            for k, v in form.errors.items():
                response_msg += f'<br />{k}: {v}'
            messages.error(request, mark_safe(response_msg))
            response = render(
                request,
                'jobs/stair_scope_row.html',
                context,
            )
            response['HX-Trigger'] = 'getmessages'
            return response
    # get
    if 'reset' in request.GET:
        return render(request, 'jobs/stair_scope_row.html', context)
    context['form'] = StairScopeRowForm(instance=stair)
    return render(request, 'jobs/stair_scope_row_edit.html', context)

def stair_scope_row_add(request):
    if request.method == 'DELETE':
        return HttpResponse()
    context = {}
    if request.method == 'POST':
        form = StairScopeRowForm(request.POST)
        if not form.errors:
            context['stair'] = form.save()
            return render(request, 'jobs/stair_scope_row.html', context)
        else:
            return _form_error_msgs(request, 'stair', form)
    if 'clone' in request.GET:
        stair = Stair.objects.get(id=request.GET['clone'])
        form = StairScopeRowForm(instance=stair)
    elif 'submittal' not in request.GET:
        raise BadRequest('submittal or clone is required')
    else:
        form = StairScopeRowForm(initial={'submittal': request.GET['submittal']})
    context['form'] = form
    return render(request, 'jobs/stair_scope_row_edit.html', context)

def sub_calendar(request, job_pk):
    job = Job.objects.get(pk=job_pk)
    xmtls = job.transmittal_set.all()
    return render(request, 'jobs/sub_calendar.html', {'xmtls': xmtls})

def transmittal_row(request, pk):
    if request.method == 'DELETE':
        return _delete_msg(request, Transmittal.objects.get(pk=pk))
    if request.method == 'POST':
        if pk != 0:
            xmtl = Transmittal.objects.get(pk=pk)
            form = TransmittalForm(request.POST, instance=xmtl)
        else:
            form = TransmittalForm(request.POST)
        from pprint import pprint
        pprint(form)
        pprint(form.fields)
        pprint(request.POST)
        if not form.errors:
            return render(request, 'jobs/transmittal_row.html', dict(xmtl=form.save()))
        else:
            return _form_error_msgs(request, 'transmittal', form)
    if 'cancel' in request.GET:
        return HttpResponse()
    if 'add' in request.GET:
        job = Job.objects.get(pk=pk)
        form = make_transmittal_form(job)
        context = dict(form=form, job=job)
        return render(request, 'jobs/transmittal_row_edit.html', context)
    xmtl = Transmittal.objects.get(pk=pk)
    if 'edit' in request.GET:
        form = TransmittalForm(instance=xmtl)
        job = xmtl.job
        if request.GET['edit'] == 'clone':
            form = TransmittalForm(initial=dict(
                job=xmtl.job,
                description=xmtl.description,
                stairs=[s.pk for s in xmtl.stairs.all()],
            ))
            xmtl = None
        form.fields['stairs'].choices = [(s.pk, s.unique_name()) for s in job.get_stairs()]
        context = dict(xmtl=xmtl, form=form, job=job)
        return render(request, 'jobs/transmittal_row_edit.html', context)
    else:
        return render(request, 'jobs/transmittal_row.html', {'xmtl': xmtl})

def current_messages(request):
    return render(request, '_messages.html', {'messages': messages.get_messages(request)})
